/*
 * Simon.c
 * Implementation of NSA Simon Block Cipher
 * Copyright 2017 Michael Calvin McCoy
 * calvin.mccoy@gmail.com
 *  # The MIT License (MIT) - see LICENSE.md
 */

#include <stdint.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include "Simon.h"

// Cipher Operation Macros
#define shift_one(x_word) (((x_word) << 1) | ((x_word) >> (word_size - 1)))
#define shift_eight(x_word) (((x_word) << 8) | ((x_word) >> (word_size - 8)))
#define shift_two(x_word) (((x_word) << 2) | ((x_word) >> (word_size - 2)))

#define rshift_three(x) (((x) >> 3) | (((x) & 0x7) << (word_size - 3)))
#define rshift_one(x)   (((x) >> 1) | (((x) & 0x1) << (word_size - 1)))

uint64_t z_arrays[5] = {
	0x19C3522FB386A45FULL, 0x16864FB8AD0C9F71ULL, 0x3369F885192C0EF5ULL,
	0x3C2CE51207A635DBULL, 0x3DC94C3A046D678BULL};

// Valid Cipher Parameters
const uint8_t simon_rounds = 72;
const uint8_t simon_block_sizes = 128;
const uint16_t simon_key_sizes = 256;
const uint8_t z_assign = 4;

uint8_t Simon_Init(Simon_Cipher *cipher_object,
	 void *key
) {

	uint8_t word_size;
	uint8_t word_bytes;
	uint16_t key_words;
	uint64_t sub_keys[4];
	uint64_t mod_mask;
	uint64_t tmp1, tmp2;
	uint64_t c;
	int i, j;

	cipher_object->block_size = simon_block_sizes;
	cipher_object->key_size = simon_key_sizes;
	cipher_object->round_limit = simon_rounds;

	cipher_object->z_seq = z_assign;
	word_size = simon_block_sizes >> 1;
	word_bytes = word_size >> 3;
	key_words = simon_key_sizes / word_size;

	mod_mask = ULLONG_MAX >> (64 - word_size);

	// Setup
	for (i = 0; i < key_words; i++) {
		memcpy(&sub_keys[i], (unsigned char *)key + (word_bytes * i),
			word_bytes);
	}

	c = 0xFFFFFFFFFFFFFFFCULL;

	// Store First Key Schedule Entry
	memcpy(cipher_object->key_schedule, &sub_keys[0], word_bytes);

	for (i = 0; i < simon_rounds - 1; i++) {
		tmp1 = rshift_three(sub_keys[key_words - 1]);

		if (key_words == 4) {
			tmp1 ^= sub_keys[1];
		}

		tmp2 = rshift_one(tmp1);
		tmp1 ^= sub_keys[0];
		tmp1 ^= tmp2;

		tmp2 = c ^ ((z_arrays[cipher_object->z_seq] >> (i % 62)) & 1);

		tmp1 ^= tmp2;

		// Shift Sub Words
		for (j = 0; j < (key_words - 1); j++) {
			sub_keys[j] = sub_keys[j + 1];
		}
		sub_keys[key_words - 1] = tmp1 & mod_mask;

		// Append sub key to key schedule
		memcpy(cipher_object->key_schedule + (word_bytes * (i + 1)),
			&sub_keys[0], word_bytes);

	}


	return 0;
}

void Simon_Encrypt_128(const uint8_t round_limit, const void *key_schedule,
	 uint8_t *plaintext) {
	uint64_t *round_key_ptr;
	uint8_t i;
	const uint8_t word_size = 64;
	uint64_t *y_word = (uint64_t*)plaintext;
	uint64_t *x_word = (((uint64_t*)plaintext) + 1);
	round_key_ptr = (uint64_t*)key_schedule;

	for (i = 0; i < round_limit; i++) {

		// Shift, AND , XOR ops
		uint64_t temp = (shift_one(*x_word) & shift_eight(*x_word))
			^ *y_word ^ shift_two(*x_word);

		// Feistel Cross
		*y_word = *x_word;

		// XOR with Round Key
		*x_word = temp ^ *(round_key_ptr + i);
	}
}

void Simon_Decrypt_128(const uint8_t round_limit, const void *key_schedule,
	 uint8_t *ciphertext) {

	const uint8_t word_size = 64;
	int8_t i;
	uint64_t *x_word = (uint64_t*)ciphertext;
	uint64_t *y_word = ((uint64_t*)ciphertext) + 1;
	uint64_t *round_key_ptr = (uint64_t*)key_schedule;


	for (i = round_limit - 1; i >= 0; i--) {

		// Shift, AND , XOR ops
		uint64_t temp = (shift_one(*x_word) & shift_eight(*x_word))
			^ *y_word ^ shift_two(*x_word);

		// Feistel Cross
		*y_word = *x_word;

		// XOR with Round Key
		*x_word = temp ^ *(round_key_ptr + i);
	}
}
