/*
 * Speck.c
 * Implementation of NSA Speck Block Cipher
 * Copyright 2017 Michael Calvin McCoy
 * calvin.mccoy@gmail.com
 *  # The MIT License (MIT) - see LICENSE.md see LICENSE.md
 */

#include <stdint.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include "Speck.h"

#define rotate_left(x,n) (((x) >> (word_size - (n))) | ((x) << (n)))
#define rotate_right(x,n) (((x) << (word_size - (n))) | ((x) >> (n)))

const uint8_t speck_rounds = 34;
const uint8_t speck_block_sizes = 128;
const uint16_t speck_key_sizes = 256;

uint8_t Speck_Init(Speck_Cipher *cipher_object,
 void *key	) {

	uint8_t word_size;
	uint8_t word_bytes;
	uint16_t key_words;
	uint64_t sub_keys[4];
	uint64_t mod_mask;
	uint64_t i;
	uint64_t tmp, tmp2;
	int j, n;

	cipher_object->block_size = speck_block_sizes;
	cipher_object->key_size = speck_key_sizes;
	cipher_object->round_limit = speck_rounds;


	word_size = speck_block_sizes >> 1;
	word_bytes = word_size >> 3;
	key_words = speck_key_sizes / word_size;

	mod_mask = ULLONG_MAX >> (64 - word_size);

	cipher_object->alpha = 8;
	cipher_object->beta = 3;

	// Setup
	for (n = 0; n < key_words; n++) {
		memcpy(&sub_keys[n], (unsigned char *)(key) + (word_bytes * n),
			word_bytes);
	}

	// Store First Key Schedule Entry
	memcpy(cipher_object->key_schedule, &sub_keys[0], word_bytes);

	for (i = 0; i < speck_rounds - 1; i++) {

		// Run Speck Cipher Round Function
		tmp = (rotate_right(sub_keys[1], cipher_object->alpha)) & mod_mask;
		tmp = (tmp + sub_keys[0]) & mod_mask;
		tmp = tmp ^ i;
		tmp2 = (rotate_left(sub_keys[0], cipher_object->beta)) & mod_mask;
		tmp2 = tmp2 ^ tmp;
		sub_keys[0] = tmp2;

		// Shift Key Schedule Subword
		if (key_words != 2) {
			// Shift Sub Words
			for (j = 1; j < (key_words - 1); j++) {
				sub_keys[j] = sub_keys[j + 1];
			}
		}
		sub_keys[key_words - 1] = tmp;

		// Append sub key to key schedule
		memcpy(cipher_object->key_schedule +
			(word_bytes * ((unsigned char)(i) + 1)), &sub_keys[0], word_bytes);
	}


	return 0;
}



void Speck_Encrypt_128(const uint8_t round_limit, const uint8_t *key_schedule,
	 uint8_t *plaintext) {

	const uint8_t word_size = 64;
	uint8_t i;
	uint64_t *y_word = (uint64_t*)plaintext;
	uint64_t *x_word = ((uint64_t*)plaintext) + 1;
	uint64_t *round_key_ptr = (uint64_t*)key_schedule;


	for (i = 0; i < round_limit; i++) {
		*x_word = ((rotate_right(*x_word, 8)) + *y_word) ^ *(round_key_ptr + i);
		*y_word = (rotate_left(*y_word, 3)) ^ *x_word;
	}
}



void Speck_Decrypt_128(const uint8_t round_limit, const uint8_t *key_schedule,
	uint8_t *ciphertext) {

	const uint8_t word_size = 64;
	int8_t i;
	uint64_t *y_word = (uint64_t*)ciphertext;
	uint64_t *x_word = ((uint64_t*)ciphertext) + 1;
	uint64_t *round_key_ptr = (uint64_t*)key_schedule;


	for (i = round_limit - 1; i >= 0; i--) {
		*y_word = rotate_right((*y_word ^ *x_word), 3);
		*x_word =
			rotate_left((uint64_t)((*x_word ^ *(round_key_ptr + i)) -
			*y_word), 8);
	}
}
